
!************************************************************************************************************************
!                                                                                                                       !
!  PROGRAM: BIG                                                                                                         !
!  This program is used to compute the Bayes factor of inquality constrained hypotheses.
!  I provide the code such that everyone can verify for themselves.                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                                                                                       !
!************************************************************************************************************************

  include 'link_f90_static.h'   ! Compiler V9
! include 'link_fnl_static.h'   ! Compiler V10 		

program BIG

    USE IMSL_LIBRARIES
    
    ! define variable
    implicit none
    ! integer
    integer                             :: i, ii, j, k, s, cout1, cout2, time, tellerline, ISEED
    integer								:: ios, numR, length, N, discard, U
    integer, dimension(:), allocatable  :: numf, numc

    ! reals/doubles
    real                                :: p, gibbsa, GG, bf, bfc, LO, UP, sd, sdc, totalf, totalc, temp
    real, dimension(:), allocatable     :: parmeans, x, y, means, ff, cc, large1, small1, large2, small2, initial, LOWER, UPPER
    real, dimension(:), allocatable     :: bfsample, bfcsample, TTf, TTc, fsam, csam, large, small
    real, dimension(:,:), allocatable   :: cov, R, B, C, IE, invcov, invB, fsample, csample
    character*256                       :: line


Open(unit = 10, file = 'Input.txt', status ='OLD', iostat=ios)
Open(unit = 20, file = 'Output.txt', status = 'replace')

tellerline = 0

! provide seed for random number generation
ISEED=12345

read(10,*)
tellerline = tellerline + 1

! read the number of parameters and constraints
read(10,*) length, numR
tellerline = tellerline + 1

! allocate, "parmeans" denotes estimates of parameters and "cov" denotes the covariance matrix 
allocate (parmeans(length+1))
allocate (cov(length,length),R(numR, length+1))

! read estimates of parameters
read(10,*)
tellerline = tellerline + 1
read(10,*) parmeans(1:length)
tellerline = tellerline + 1
parmeans(length+1) = -1
! read the covairance matrix
read(10,*)
tellerline = tellerline + 1
do i = 1,length
   read(10,*) cov(i,1:length) 
   tellerline = tellerline + 1
end do
 
! read the restriction matrix (R-r)
read(10,*)
tellerline = tellerline + 1

do i = 1, numR
   read(10,*) R(i,1:(length+1)) 
   tellerline = tellerline + 1	
end do

close(10)

call RNSET(ISEED)
! U is the number of iteration for computing MC error
U=1000
    ! allocate variable and matrix needed in the program
    allocate (x(length+1),y(length+1),means(length+1),initial(length+1),LOWER(length+1),UPPER(length+1),large(numR), small(numR))
    allocate (ff(numR),cc(numR),large1(numR),large2(numR),small1(numR),small2(numR),numf(numR),numc(numR))
    allocate (bfsample(U),bfcsample(U),TTf(U),TTc(U),fsam(U),csam(U))
    allocate (IE(length,length),B(length,length),C(length,length))
    allocate (invcov(length,length),invB(length,length))
    allocate (fsample(numR,U),csample(numR,U))

    B=0
    IE=0
    ! calculate the matrix used for the conditional distribution, that is, C=I-[diag(cov^(-1))]^(-1)cov^(-1)
    call LINRG(cov,invcov)
    do i=1,length
       B(i,i)=invcov(i,i)
       IE(i,i)=1
    end do
    call LINRG(B,invB)
    do i=1,length
       do j=1,length
          C(i,j)=IE(i,j)-invB(i,i)*invcov(i,j)
       end do
    end do  
     
    initial=parmeans  
      
    
    do ii=1,numR ! decompose the Bayes factor, "ii" denotes each line of the restriction matrix (R-r)     
    ! define initial values that are in agreement with the first ii constraints.
     do i=1,length 
          do k=1,ii
             if (R(k,i)>0) then
             ! the parameter should be larger than
                large(k)=(R(k,i)*initial(i)-sum(R(k,1:(length+1))*initial))/R(k,i)
             else
                large(k)=-10**20
             end if
             if (R(k,i)<0) then
             ! the parameter should be smaller than
                small(k)=(R(k,i)*initial(i)-sum(R(k,1:(length+1))*initial))/R(k,i)
             else
                small(k)=10**20
             end if
          end do
          if (ii<numR) then
          ! the last "numR"-"ii" constraints are not used
          do s=(ii+1),numR
             large(s)=-10**20
             small(s)=10**20
          end do
          end if
          LOWER(i)=MAXVAL(large)
          UPPER(i)=MINVAL(small) 
          ! this parameter does not have constraint
          if (LOWER(i)==-10**20.AND.UPPER(i)==10**20) then
          initial(i)=parmeans(i)
          ! this parameter should be larger than "LOWER"
          else if (UPPER(i)==10**20) then
          initial(i)=LOWER(i)+0.1
          ! this parameter should be smaller than "UPPER"
          else if (LOWER(i)==-10**20)then
          initial(i)=UPPER(i)-0.1
          ! this parameter should be larger than "LOWER" and smaller than "UPPER"
          else
          initial(i)=LOWER(i)+(UPPER(i)-LOWER(i))/2
       end if
     end do
       
     ! Gibbs sampling iteration 
       N=1000
     ! "discard" denotes the burn-in period for each element of the decomposed Bayes factor   
       discard=100     
     ! 777 is for the computation of the fit, that is, the posterior proportion     
  777  cout1=0
       x=initial
       large1=0
       small1=0
       do time=1,N+discard ! times of Gibbs sampler is N
          means=0  
          do i=1,length
          ! each iteration with constrained sampling
          ! "means" is the mean of conditional posterior normal distribution of theta_(i)
             means(i)=parmeans(i)+sum(C(i,1:length)*(x(1:length)-parmeans(1:length)))-C(i,i)*(x(i)-parmeans(i))
             
             if(ii > 1) then ! except the first line of the restriction matrix
                do k=1,(ii-1)
                   if (R(ii-k,i)>0) then
                   ! the parameter should be larger than
                      large1(k)=(R(ii-k,i)*x(i)-sum(R(ii-k,1:(length+1))*x))/R(ii-k,i)
                   else
                      large1(k)=-10**20
                   end if
                   if (R(ii-k,i)<0) then
                   ! the parameter should be smaller than
                      small1(k)=(R(ii-k,i)*x(i)-sum(R(ii-k,1:(length+1))*x))/R(ii-k,i)
                   else
                      small1(k)=10**20
                   end if
                end do
                do s=ii,numR
                   large1(s)=-10**20
                   small1(s)=10**20
                end do
                ! compute the Lower and upper bound, the parameter should be sampled in that interval
                LO=(MAXVAL(large1)-means(i))/sqrt(1/invcov(i,i))
                UP=(MINVAL(small1)-means(i))/sqrt(1/invcov(i,i))
                GG=RNUNF()
                p=ANORDF(LO)+GG*(ANORDF(UP)-ANORDF(LO))
                !avoid p=1 and p=0
                if (p>0.999999)then
                p=0.999999
                end if
                if (p<0.000001)then
                p=0.000001
                end if
                gibbsa=ANORIN(p)
                x(i)=gibbsa*sqrt(1/invcov(i,i))+means(i)
             else
                GG=RNNOF()
		        x(i)=GG*sqrt(1/invcov(i,i))+means(i)
             end if
          end do
          ! whether satisfy the constraint
          ! posterior/fit       
          if(sum(R(ii,1:(length+1))*x)>0 .AND. time>discard) then
             cout1=cout1+1
          end if
       end do
       
       ff(ii)=real(cout1)/real(N)
       ! this is to determine the number of iterations based on the value of fit
       if (ff(ii)<0.5 .AND. N==1000) then
       N=3000
       go to 777
       else if (ff(ii)<0.166 .AND. N==3000) then
       N=9600
       go to 777
       else if (ff(ii)<0.042 .AND. N==9600) then
       N=120000
       go to 777
       else if (ff(ii)<0.008 .AND. N==120000) then
       N=360000
       go to 777
       else if (ff(ii)<0.0014 .AND. N==360000) then 
       N=2520000
       go to 777
       else if (ff(ii)<0.0002 .AND. N==2520000) then  
       N=10000000
       go to 777
       end if
       numf(ii)=N
       
       ! generate the sample of fit for computing the MC error 
       Call RNBET(real(cout1+1),real(N-cout1+1),TTf)
           fsample(ii,1:U)=TTf
           
       ! monitor the process
       print '("Constraint", I4 " for    fit     has been evaluated", I10)', ii
  
       N=1000
       ! 888 is for the computation of the complexity, that is, the prior proportion 
  888  cout2=0
       y=initial
       large2=0
       small2=0
       do time=1,N+discard ! times of Gibbs sampler is N
          means=0
          do i=1,length
          ! each iteration with constrained sampling
             
             if(ii > 1) then ! except the first line of the restriction matrix
                do k=1,(ii-1)
                   if (R(ii-k,i)>0) then
                   ! the parameter should be larger than
                      large2(k)=(R(ii-k,i)*y(i)-sum(R(ii-k,1:(length+1))*y))/R(ii-k,i)
                   else
                      large2(k)=-10**20
                   end if
                   if (R(ii-k,i)<0) then
                   ! the parameter should be smaller than
                      small2(k)=(R(ii-k,i)*y(i)-sum(R(ii-k,1:(length+1))*y))/R(ii-k,i)
                   else
                      small2(k)=10**20
                   end if
                end do
                do s=ii,numR
                   large2(s)=-10**20
                   small2(s)=10**20
                end do
                ! compute the lower and upper bound, then parameter should be sampled in that interval
                !prior
                LO=MAXVAL(large2)/(10**20)
                UP=MINVAL(small2)/(10**20)
                GG=RNUNF()
                p=ANORDF(LO)+GG*(ANORDF(UP)-ANORDF(LO))
                !avoid p=1 and p=0
                if (p>0.999999)then
                p=0.999999
                end if
                if (p<0.000001)then
                p=0.000001
                end if
                gibbsa=ANORIN(p)
                y(i)=gibbsa*(10**20)
             else
		        GG=RNNOF()
		        y(i)=GG*(10**20)
             end if  
          end do
          ! whether satisfy the constraint
          ! prior/complexity
          if(sum(R(ii,1:(length+1))*y)>0 .AND. time>discard) then
             cout2=cout2+1
          end if
       end do
       cc(ii)=real(cout2)/real(N)
       ! this is to determine the number of iterations based on the value of complexity
       if (cc(ii)<0.5 .AND. N==1000) then
       N=3000
       go to 888
       else if (cc(ii)<0.166 .AND. N==3000) then
       N=9600
       go to 888
       else if (cc(ii)<0.042 .AND. N==9600) then
       N=120000
       go to 888
       else if (cc(ii)<0.008 .AND. N==120000) then
       N=360000
       go to 888
       else if (cc(ii)<0.0014 .AND. N==360000) then 
       N=2520000
       go to 888
       else if (cc(ii)<0.0002 .AND. N==2520000) then  
       N=10000000
       go to 888
       end if
       numc(ii)=N
       ! generate the sample of fit for computing the MC error 
    Call RNBET(real(cout2+1),real(N-cout2+1),TTc)
    csample(ii,1:U)=TTc
    
    ! monitor the process
    print '("Constraint", I4 " for complexity has been evaluated")' , ii
    
    end do
    
totalf=1
totalc=1
bf=1
! multiple all the fits and complexities 
do i = 1,numR
   totalf=totalf*ff(i)
   totalc=totalc*cc(i)
   bf=bf*(ff(i)/cc(i))
end do   

! compute the bayes factor of the inequality constrained hypothesis versus its complement
bfc=bf/((1-totalf)/(1-totalc))

! compute the sample of the bayes factor BF_ia and BF_ic, which is used to compute MC error
do j = 1,U
   bfsample(j)=1
   bfcsample(j)=1
   fsam(j)=1
   csam(j)=1
   do i = 1,numR
      bfsample(j)=bfsample(j)*fsample(i,j)/csample(i,j)
      fsam(j)=fsam(j)*fsample(i,j)
      csam(j)=csam(j)*csample(i,j)
   end do
bfcsample(j)=fsam(j)/csam(j)/((1-fsam(j))/(1-csam(j)))   
end do

! compute the 95% credible interval
do i = 1,U
do j=1,(U-i)
   if (bfsample(j)>bfsample(j+1)) then
      temp=bfsample(j)
      bfsample(j)=bfsample(j+1)
      bfsample(j+1)=temp
   end if 
   if (bfcsample(j)>bfcsample(j+1)) then
      temp=bfcsample(j)
      bfcsample(j)=bfcsample(j+1)
      bfcsample(j+1)=temp
   end if 
end do
end do      

! compute the MC error (standard deviation)
sd=sqrt(sum((sum(bfsample)/U-bfsample(1:U))**2)/U)
sdc=sqrt(sum((sum(bfcsample)/U-bfcsample(1:U))**2)/U)

! the results
write(20,'(A)') 'Result:'
write(20,*)
write(20,'(A)') '   Fits        numbers of iterations'
do i = 1, numR
   write(20,'(f8.3,I15)') ff(i),  numf(i)
end do
write(20,*)
write(20,'(A)') 'Complexities   numbers of iterations'
do i = 1, numR
   write(20,'(f8.3,I15)') cc(i),  numc(i) 
end do
write(20,*)
write(20,'(A)') 'Total fit      Total complexity'
write(20,'(f8.3,f15.3)') totalf, totalc
write(20,*)
write(20,'(A)') '   BFia   MC error (standard deviation)   2.5percentile  97.5percentile'
write(20,'(f8.3,f13.3,f30.3,f16.3)') bf, sd, bfsample(U*2.5/100),bfsample(U*97.5/100)
write(20,*)
write(20,'(A)') '   BFic   MC error (standard deviation)   2.5percentile  97.5percentile'
if (totalf .EQ.1) then
write(20,'(A)') '   infinit'
else
write(20,'(f8.3,f13.3,f30.3,f16.3)') bfc, sdc, bfcsample(U*2.5/100),bfcsample(U*97.5/100)
end if

close(20)     
end program BIG
